/**
 * Last Modified by: $Author: rjeffers $
 * On: $Date: 2005/12/12 21:05:09 $
 */
 
package kaos.core.csi;

import java.util.Iterator;
import java.util.HashMap;
import java.io.Serializable;

import kaos.core.util.NamespaceValidator;
import kaos.ontology.vocabulary.ActionConcepts;
import kaos.ontology.util.ConceptTranslator;

import java.net.URISyntaxException;

/**
 *   This class containes information describing an Action instance.
 *   
 *   $Revision: 1.13 $
 */
public class ActionInstanceDescriptionImpl extends OntInstanceDescriptionImpl implements ActionInstanceDescription, Cloneable , Serializable 
{
    /**
     * Comment for <code>serialVersionUID</code>
     */
    private static final long serialVersionUID = -5894865716501809678L;
    
    private String _conceptMappingName = null;
    private String _propertyPopulatorKey = null;
    
    /**
     * Create an instance of the ActionInstanceDescription described by the given arguments.
     *
     * @param  actionId             String containing the id of the given action.
     *
     * @param  actionClassName    	String containing the ontological name of the given action.
     * 
     * @param  actor              	String containing the id of the actor performing the action.
     *                               
     * @throws                      URISyntaxException if the argument to create an URI is syntactically incorrect .
     * 
     */
    public ActionInstanceDescriptionImpl (String actionId, String actionClassName, String actor) throws URISyntaxException
    {
        super(actionId, actionClassName);
        
        if ( actor == null ) {
        	throw new URISyntaxException(actor,"Actor id is null!");
        }
        actor = NamespaceValidator.validateURI(actor);
        OntPropertyDescription performedByProperty = new OntPropertyDescriptionImpl(ActionConcepts.performedBy);
        performedByProperty.addValue(actor);
        
        addProperty(performedByProperty);
        
        _conceptMappingName = ConceptTranslator.KAOS_MAPPING;
    }
	
    /**
     * Create an instance of the ActionInstanceDescription described by the given arguments.
     *
     * @param actionId             	String containing the id of the given action.
     *
     * @param actionClassName       String containing the ontological name of the given action.
     * 
     * @param actor              	String containing the id of the actor performing the action.
     * 
     * @param properties            HashMap containing the ontological names and their values of properties 
     *                              of the given action.
     *                               
     * @throws                      URISyntaxException if the argument to create an URI is syntactically incorrect .
     * 
     */
    public ActionInstanceDescriptionImpl (String actionId, 
    									  String actionClassName, 
    									  String actor, 
    									  HashMap properties)
    throws URISyntaxException
    {
        super(actionId, actionClassName);
       
        if ( actor == null ) {
        	throw new URISyntaxException(actor,"Actor id is null!");
        }
        
        if ( properties == null ) {
        	System.out.println("Action properties are null!  Unable to create a valid instance of ActionInstanceDescription!");
        	return;
        }
        
        NamespaceValidator.validateURI(actor);
        OntPropertyDescription performedByProperty = new OntPropertyDescriptionImpl(ActionConcepts.performedBy);
        performedByProperty.addValue(actor);
        addProperty(performedByProperty);
        
        for(Iterator propIT = properties.keySet().iterator(); propIT.hasNext(); )
        {
            String propertyName = (String) propIT.next();
		    Object value = properties.get(propertyName);
		    OntPropertyDescription propDesc = new OntPropertyDescriptionImpl(propertyName);
		    if ( value instanceof String) {
		    	propDesc.addValue((String)value);
		    }
		    else {
		    	propDesc.setValueDescForValue( getInstanceID(), value);	    
		    }   
			addProperty(propDesc);
        }
        _conceptMappingName = ConceptTranslator.KAOS_MAPPING;
    }
    
    /**
     * Create an instance of the ActionInstanceDescription described by the given OWL String.
     *
     * @param  owlString            String the action id, class name, actor id and, optionally, properties.
     *                               
     * @throws                      URISyntaxException if the argument to create an URI is syntactically incorrect .
     * 
     */
    public ActionInstanceDescriptionImpl (String owlString)
    throws URISyntaxException
    {
    	//TODO Parse the OWL String to retrieve the action id, class name, properties (if present) and actor id.
    	super(null, "No class name");
    	_conceptMappingName = ConceptTranslator.KAOS_MAPPING;
    }
    
    /* 
     * @see kaos.core.csi.ActionInstanceDescription#getActor()
     */
    public String getActor ()
    {
        return (String)getProperty(ActionConcepts.performedBy()).getValues().toArray()[0];
    }
    
    /* 
     * @see kaos.core.csi.ActionInstanceDescription#getActionClassName()
     */
    public String getActionClassName ()
    {
        return (String)getTypes().toArray()[0];
    }
    
    public void setConceptMappingName(String name) 
    {
        _conceptMappingName = name;
    }
    
     /* 
     * @see kaos.core.csi.ActionInstanceDescription#getConceptMappingName()
     */
    public String getConceptMappingName ()
    {
        return _conceptMappingName;
    }
    
    /* 
     * @see kaos.core.csi.ActionInstanceDescription#getPropertyPopulatorKey()
     */
    public String getPropertyPopulatorKey()
    {
        return _propertyPopulatorKey;
    }
    
    /* 
     * @see kaos.core.csi.ActionInstanceDescription#setPropertyPopulatorKey(java.lang.String)
     */
    public void setPropertyPopulatorKey(String aKey)
    {
        _propertyPopulatorKey = aKey;
    }
    
    /* 
     * @see kaos.core.csi.ActionInstanceDescription#toString()
     */
    public String toString()
    {
       StringBuffer buffer = new StringBuffer();
	   buffer.append("\nAction Instance Description:" +
					 "\n\tid: " + getName());
       buffer.append("\n\tclass: " + getActionClassName());
       buffer.append("\n\tperformed by: " + getActor());
       buffer.append("\n\tconceptMapping:");
       buffer.append(_conceptMappingName + "\n");
       buffer.append("\n\tproperties: ");
       
       for(Iterator it = getProperties().iterator(); it.hasNext(); )
       {
           OntPropertyDescription currentProperty = (OntPropertyDescription)it.next();
           
           if (currentProperty.getName().equals(ActionConcepts.performedBy()))
               continue;
           
		   buffer.append("\n\t\t" + currentProperty + "\n");
       }
       
       
  
       return buffer.toString();
    }
    
    public boolean equals (Object o) {
        return super.equals(o);
    }
    
    public int hashCode() {
        return super.hashCode();
    }
    
    /* 
     * Implementation of the clone method
     */
    public  Object clone()
	{	
		return (ActionInstanceDescriptionImpl) super.clone();
	}
}