/**
 * Last Modified by: $Author: mjohnson $
 * On: $Date: 2006/01/24 15:16:46 $
 */

/* This class represents an ontology instance as a named set of properties. 

   This class is actually nothing more than a hashtable which stores
   KAoSProperty objects hashed on their names.
 */

package kaos.core.csi;

import java.rmi.dgc.VMID;
import java.util.Iterator;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.io.Serializable;

import java.net.URI;
import java.net.URISyntaxException;

import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.NodeIterator;
import com.hp.hpl.jena.ontology.DatatypeProperty;
import com.hp.hpl.jena.ontology.Individual;
import com.hp.hpl.jena.ontology.ObjectProperty;
import com.hp.hpl.jena.ontology.OntClass;

import kaos.ontology.util.SerializableOntModelImpl;
import kaos.ontology.vocabulary.DefaultNamespaces;
import kaos.ontology.vocabulary.OWLConcepts;
import kaos.ontology.vocabulary.RDFConcepts;

/**
 *  Represents an instance of an ontology class.
 *  Each property's value may have description, which can be another OntInstanceDescriptionImpl thus allowing 
 *  for unlimited nesting of OntInstanceDescriptionImpl's.
 *  
 *  @see OntInstanceDescription
 *   
 *  $Revision: 1.5 $
 */
public class OntInstanceDescriptionImpl implements OntInstanceDescription, Cloneable, Serializable
{
    /**
     * Comment for <code>serialVersionUID</code>
     */
    private static final long serialVersionUID = 1L;

    private String instanceNameURI;
    private Set myTypes;
    private Map myProperties;
    
    /**
     * Basic constructor
     * 
     * @param instanceName - instance name
     * 
     * @param className - initial class of instance
     * 
     * @throws URISyntaxException when the class or instance name are not URI
     */
    public OntInstanceDescriptionImpl(String instanceName, String className) throws URISyntaxException
    {
        super();
        
        if ( instanceName == null )
            instanceName = getInstanceID();
        
        if ( className == null )
            throw new URISyntaxException(className, "Class name is null!");
        
        URI testInstanceName = new URI(instanceName);
        
        URI testclassName = new URI(className);
        
        instanceNameURI = instanceName;
        myTypes = new HashSet();
        myTypes.add(className);
        myProperties = new HashMap();
    }
    
    /**
     * Constructor creating an instance based on Jena model containg OWL defining this instance
     * 
     * @param instanceModel - instance model
     */
    public OntInstanceDescriptionImpl(SerializableOntModelImpl instanceModel) throws URISyntaxException
    {
        super();
        myTypes = new HashSet();
        myProperties = new HashMap();
        
        instanceNameURI =  instanceModel.getOntologyStructureId();
        
        if (instanceNameURI == null)
            throw new URISyntaxException("", "Instance name is null!");
        
        if (instanceNameURI == "")
            throw new URISyntaxException("", "Instance name is empty!");
        
        //TODO: test if instanceNameURI is not URI
        
        // Get instance from model
        Individual instance = instanceModel.getIndividual(instanceNameURI);
        
        if (instance == null)
            throw new URISyntaxException(instanceNameURI, "Instance name is not valid!");
        
        // Get types of the instance
        for (Iterator instanceTypeIT = instance.listRDFTypes(true); instanceTypeIT.hasNext(); )
        {
            Resource currentClsRes = (Resource) instanceTypeIT.next();
            OntClass currentCls = (OntClass) currentClsRes.as( OntClass.class );
            
            if (currentCls.getURI().equals(OWLConcepts.Thing()))
                continue;
            
            myTypes.add(currentCls.getURI());
        }
        
        // Get properties of the instance
        for (Iterator propertiesIT = instance.listProperties(); propertiesIT.hasNext(); )
        {
            Statement s = (Statement)propertiesIT.next(); 
            String predicateURI = s.getPredicate().getURI();
            if (!predicateURI.equals(RDFConcepts._type_))
            {
                OntPropertyDescription currentProperty = null;
                
                if (s.getPredicate().canAs(ObjectProperty.class))
                {
                    ObjectProperty currentObjectProperty = (ObjectProperty)s.getPredicate().as(ObjectProperty.class);
                    
                    currentProperty = new OntPropertyDescriptionImpl(currentObjectProperty.getURI());
                    
                    for (NodeIterator propertyValuesIT = instance.listPropertyValues(currentObjectProperty); propertyValuesIT.hasNext(); )
                    {
                        RDFNode rdn = (RDFNode)propertyValuesIT.next();
                        String currentPropertyValue = ((Individual)rdn.as(Individual.class)).getURI();
                        
                        if(currentPropertyValue.startsWith("#"))
                        {
                            currentPropertyValue = currentPropertyValue.substring(1);
                        }
                        
                        currentProperty.addValue(currentPropertyValue);
                        
                        instanceModel.setOntologyStructureId(currentPropertyValue);
                        try 
                        {
                            OntInstanceDescription currentPropertyValueDesc = new OntInstanceDescriptionImpl(instanceModel);
                            currentProperty.setValueDescForValue(currentPropertyValue, currentPropertyValueDesc);
                        } 
                        catch (URISyntaxException e) 
                        {
                            // There is no description for this instance in this model
                        }
                        instanceModel.setOntologyStructureId(instanceNameURI);
                    }
                }
                else
                {
                    DatatypeProperty currentDatatypeProperty = (DatatypeProperty)s.getPredicate().as(DatatypeProperty.class);
                    currentProperty = new OntPropertyDescriptionImpl(currentDatatypeProperty.getURI());
                    
                    for (NodeIterator propertyValuesIT = instanceModel.listObjectsOfProperty(instance, currentDatatypeProperty); propertyValuesIT.hasNext(); )
                        currentProperty.addValue(propertyValuesIT.next().toString());
                }
                
                myProperties.put(currentProperty.getName(), currentProperty);
            }
        }        
    }

    public boolean equals (Object obj)
    {
        if (obj instanceof OntInstanceDescriptionImpl) 
        {
            OntInstanceDescriptionImpl other = (OntInstanceDescriptionImpl) obj;
            boolean test1 = other.instanceNameURI.equals(instanceNameURI);
            boolean test2 = other.myProperties.equals(myProperties);
            boolean test3 = other.myTypes.equals(myTypes);
            
            return (other.instanceNameURI.equals(instanceNameURI) &&
                    other.myProperties.equals(myProperties) &&
                    other.myTypes.equals(myTypes));
        }
        return false;
    }
    
    public int hashCode() 
    {
        return instanceNameURI.hashCode() &  myProperties.hashCode() & myTypes.hashCode();
    }
    
    /**
     * @see kaos.core.csi.OntInstanceDescription#addProperty(kaos.core.csi.OntPropertyDescription)
     */
    public void addProperty(OntPropertyDescription newProperty) 
    {
        myProperties.put(newProperty.getName(), newProperty);
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#addType(java.lang.String)
     */
    public void addType(String className) throws URISyntaxException 
    {
        myTypes.add(className);
    }

    /** 
     * @see kaos.core.csi.OntInstanceDescription#getName()
     */
    public String getName() 
    {
        return instanceNameURI;
    }
    
    /**
     * @see kaos.core.csi.OntInstanceDescription#getTypes()
     */
    public Set getTypes() 
    {    
        return myTypes;
    }
    
    /**
     * @see kaos.core.csi.OntInstanceDescription#getProperties()
     */
    public Set getProperties() 
    {
        Set myResult = new HashSet();
        
        // Check if there are any properties
        if (myProperties.values() == null)
            return myResult;
        
        // Gather properties in a set
        for (Iterator myPropertiesIt = myProperties.values().iterator(); myPropertiesIt.hasNext(); )
            myResult.add((OntPropertyDescription)myPropertiesIt.next());
        
        return myResult;
    }
    
    /**
     * @see kaos.core.csi.OntInstanceDescription#getProperty(java.lang.String)
     */
    public OntPropertyDescription getProperty(String propertyName) 
    {
        return (OntPropertyDescription)myProperties.get(propertyName);
    }	
	
    /**
     * Get the given property if nested in the train of interrelated instances.
     * If the property is not found in this instance, recursively search for the property in 
     * OntInstanceDescriptions if nested within this one.
     * 
     * @param propertyName - URI for the property to find
     * @return - the property description
     */
    public OntPropertyDescription getNestedProperty(String propertyName)
    {
        OntPropertyDescription ret = null;
        
        OntPropertyDescription currentProperty = getProperty(propertyName);
        
        if (currentProperty != null)
            return currentProperty;
        
        for(Iterator propIT = getProperties().iterator(); propIT.hasNext(); )
        {
            currentProperty = (OntPropertyDescription)propIT.next();
            
            for(Iterator propVIT = currentProperty.getValues().iterator(); propVIT.hasNext(); )
            {
                String currentValue = (String) propVIT.next();
                
                if (currentProperty.getValueDescForValue(currentValue) instanceof OntInstanceDescription)
                {
                    OntInstanceDescriptionImpl currentValueDesc = (OntInstanceDescriptionImpl)currentProperty.getValueDescForValue(currentValue);
                    
                    ret = currentValueDesc.getNestedProperty(propertyName);
                    
                    if (ret != null)
                        return ret;
                }
            }
        }
        
       return ret;
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#removeType(java.lang.String)
     */
    public void removeType(String name) 
    {
        myTypes.remove(name);
    }

    /**
     * @see kaos.core.csi.OntInstanceDescription#removeProperty(java.lang.String)
     */
    public void removeProperty(String propertyName) 
    {
        myProperties.remove(propertyName);
    }
  
    /**
     * Implementation of the clone method
     * 
     * @return     clone of the object
     */
    public  Object clone()
    {       
        OntInstanceDescriptionImpl myClone = null;               
        
        try 
        {
            myClone = (OntInstanceDescriptionImpl) super.clone();
        }
        catch (CloneNotSupportedException e) 
        {
            // this should never happen
            throw new InternalError(e.toString());
        }
        
        for(Iterator it = getProperties().iterator(); it.hasNext(); )
        {
            OntPropertyDescriptionImpl currentProperty = (OntPropertyDescriptionImpl)it.next();
            
            myClone.addProperty((OntPropertyDescription)currentProperty.clone());
        }
        
        return myClone; 
    }
    
    /**
     * 
     * @return String containing the unique ID for the instance.
     */
    protected String getInstanceID()
    {
    	return DefaultNamespaces.defaultKAoSInstancesNamespace + (new VMID()).toString();
    }
   
    /* 
     * Returns a string representation of this instance.
     */
    public String toString()
    {
       StringBuffer buffer = new StringBuffer();
	   buffer.append("\nOntology Instance Description:");
	   buffer.append("\n\tname: " + getName());
       buffer.append("\n\ttypes: " + getTypes());
       buffer.append("\n\tproperties: " + getProperties() );
       
       return buffer.toString();
    }

    /* (non-Javadoc)
     * @see kaos.core.csi.OntInstanceDescription#getInstanceOntModel()
     */
    public SerializableOntModelImpl getInstanceOntModel() 
    {
        SerializableOntModelImpl _instanceModel =  new SerializableOntModelImpl();        
        
        OntClass _genericType = _instanceModel.createClass(OWLConcepts.Thing());
        
        String baseInstanceType = (String)myTypes.toArray()[0];
        OntClass instanceType = _instanceModel.createClass(baseInstanceType); 
        Individual _instance = _instanceModel.createIndividual(instanceNameURI, instanceType); 
        
        _instanceModel.setOntologyStructureId(instanceNameURI);
        
        // Add additional types to the instance, if exists
        if (myTypes.size() > 1)
            for (Iterator myTypesIT = myTypes.iterator(); myTypesIT.hasNext(); )
            {
                String currentType = (String) myTypesIT.next();
                
                if (currentType.equals(baseInstanceType))
                    continue; // Skip the first type; already added
                
                instanceType = _instanceModel.createClass(currentType);
                _instance.addRDFType(instanceType);
            }
         
        // Add properties
        for (Iterator myPropertiesIT = myProperties.keySet().iterator(); myPropertiesIT.hasNext(); )
        {
            String currentPropertyName = (String)myPropertiesIT.next();
            OntPropertyDescription currentProperty = (OntPropertyDescription)myProperties.get(currentPropertyName);
            
            // Check if property has any value
            if (currentProperty.getValues().size() < 1 )
                continue;
            
            String firstPropertyValue = (String)currentProperty.getValues().toArray()[0];
            
            // Check if the value is URI
            if (firstPropertyValue.indexOf(':') > 0)
            {
                // This is a object property
                ObjectProperty prop = _instanceModel.createObjectProperty(currentPropertyName);
                
                for (Iterator propertyValuesIT = currentProperty.getValues().iterator(); propertyValuesIT.hasNext(); )
                {
                    String currentValue = (String)propertyValuesIT.next();
                    
                    // Check if the value is URI
                    if (currentValue.indexOf(':') <= 0)
                    {
                        // If the current value is not URI; skip it
                        // TODO: notify about the error
                        continue;
                    }
                    
                    Individual val = _instanceModel.createIndividual(currentValue, _genericType);          
                    _instance.addProperty(prop, val);
                    
                    if (currentProperty.getValueDescForValue(currentValue) != null)
                        _instanceModel.add(((OntInstanceDescription)currentProperty.getValueDescForValue(currentValue)).getInstanceOntModel());
                }
            } 
            else
            {
                // This is a datatype property
                DatatypeProperty prop = _instanceModel.createDatatypeProperty(currentPropertyName);
                
                for (Iterator propertyValuesIT = currentProperty.getValues().iterator(); propertyValuesIT.hasNext(); )       
                    _instance.addProperty(prop, propertyValuesIT.next());
            }     
        }
        
        return _instanceModel;
    }  
}