/**
 * Last Modified by: $Author: rjeffers $
 * On: $Date: 2005/12/10 05:55:11 $
 */
package kaos.core.csi.policy;

import kaos.core.csi.ActionInstanceDescription;
import kaos.core.csi.ActionInstanceDescriptionImpl;
import kaos.core.service.boot.KAoSBootstrapper;
import kaos.core.service.directory.KAoSAgentDirectoryServiceProxy;
import kaos.policy.guard.ActionPermission;
import kaos.policy.guard.GuardRetriever;
import kaos.policy.guard.Guard;
import kaos.core.util.Logger;
import kaos.core.util.PolicyMsg; 

import javax.agent.service.ServiceFailure;

import java.net.URISyntaxException;
import java.security.Permission;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashMap;
import java.util.List;
import java.util.Vector;


/**
 * Defines the operations to check permissions, get obligations, analyze policies, etc.
 * 
 * @author KAoS Team
 * $Revision: 1.17 $
 */
public class PolicyCheckingImpl 
    implements PolicyAdvice, PolicyChecking
{
    private static final PolicyCheckingImpl     __INSTANCE = new PolicyCheckingImpl() ;
    private static boolean                      __INITIALIZED = false;
        
    private KAoSAgentDirectoryServiceProxy      _dirProxy = null;
    private Logger                              _logger = Logger.getLogger(this);
    
           
    /**
     * Retrieve an instance of PolicyChecking.  The returned instance has initialized communication with the DS.
     *
     * @param transportName String specifying the name of the transport to use for communication with the KAoS Directory Service.
     * 
     * @return a singleton, PolicyChecking.
     */
    public static PolicyChecking getInstance(String transportName)
    throws Exception
    {
        if ( __INITIALIZED ) return __INSTANCE;
        
            __INSTANCE.initialize(transportName);
            __INITIALIZED = true;
            return __INSTANCE;
    }
    
    /**
     * Retrieve an instance of PolicyAdvice.  The returned instance has initialized communication with the DS.
     *
     * @param transportName String specifying the name of the transport to use for communication with the KAoS Directory Service.
     * 
     * @return a singleton, PolicyAdvice.
     */
    public static PolicyAdvice getPolAdviceInstance(String transportName)
    throws Exception
    {
        if ( __INITIALIZED ) return __INSTANCE;
        
            __INSTANCE.initialize(transportName);
            __INITIALIZED = true;
            return __INSTANCE;
    }
    
    /**
     * Retrieve an instance of PolicyChecking.  Thy returned instance may or may not be able to communicate with
     * the Directory Service.  If the communication with the Directory Service is not supported, then 
     * all requested operations will be invoked only on the local Guard.
     *    
     * @return a singleton, PolicyChecking.
     */
    public static PolicyChecking getInstance()
    throws Exception
    {
        return __INSTANCE;
    }
    
    private PolicyCheckingImpl()
    {
    }
    
    /*****************************************************************************************
       kaos.core.csi.policy.PolicyChecking interface implementation - begin
    *****************************************************************************************/
    /**
     * The method check if the given action is permitted according to the current set of policies
     *
     * @param  perm                Permission to be checked by the Guard in order to allow/disallow
     *                             an action.
     *
     * @param  context             An Object describing the context of the action.   
     * 
     * @exception                  KAoSSecurityException if the action is not allowed.
     *
     * @exception                  NullPointerException if the Permission argument is null.
     *
     */
    public void checkPermission (Permission perm, Object context) 
        throws KAoSSecurityException, NullPointerException, ServiceFailure
    {
        Guard guard = GuardRetriever.getGuard();
        if ( guard != null ) {
            guard.checkPermission(perm,context);
            return;
        }
        if ( _dirProxy != null ) {
            _dirProxy.checkPermission(perm,context);
            return;
        }
       throw new ServiceFailure("Policy service unavailable! Unable to check permission for: " + perm);
    }
    
    /**
     * The method check if the given action is permitted according to the current set of policies.
     *
     * @param  actorID              String containing the id of the policy actor.
     *
     * @param attemptedActionName   String containing the ontological name of the attempted action.
     *
     * @param actionProperties      HashMap containing the ontological names and their values of properties 
     *                              of the attempted action.
     * 
     * @throws                      KAoSSecurityException if the action is not allowed.
     *
     * @throws                      NullPointerException if the Permission argument is null.
     *
     * @throws                      ServiceFailure if the policy service is not available.
     */
    public void checkPermission (String actorID,
                                 String attemptedActionName,
                                 HashMap actionProperties) 
        throws KAoSSecurityException, NullPointerException, ServiceFailure
    {
        ActionInstanceDescription aid = getActionInstanceDescription(actorID,
                                                                     attemptedActionName,
                                                                     actionProperties);
        if ( aid == null ) throw new NullPointerException("One of the arguments is null. Actor: " 
                                                           + actorID + "\n    ActionName:       " + attemptedActionName
	                                                       +           "\n    ActionProperties: " + actionProperties);                                                           
        
        checkPermission(new ActionPermission(actorID, aid), null);
    }

    /**
     * The method check if the given action is permitted according to the current set of policies.
     *
     * @param  aid					ActionInstanceDescription describing action
     *
     * @param pObserver        		PolicyDecisionObserver, an entity interested in receiving updates
	 *  							whenever policy decisions change.
	 *  
     * @throws                      KAoSSecurityException if the action is not allowed.
     *
     * @throws                      NullPointerException if the Permission argument is null.
     *
     * @throws                      ServiceFailure if the policy service is not available.
     */
    public void checkPermission (ActionInstanceDescription aid,
    							 PolicyDecisionObserver pObserver) 
        throws KAoSSecurityException, NullPointerException, ServiceFailure
	{
        if ( aid == null ) throw new NullPointerException("The ActionInstanceDescription is null.");                                                           
        
        String actorID = (String)aid.getActor();
		checkPermission(new ActionPermission(actorID, aid), null);		
	}
	
    /**
     * The method checks if the given action is permitted according to the current set of policies.
     * However, if the given action is not permitted by a policy, instead of
     * stopping there, it goes on to find all those policies that forbid the
     * given action, and then returns a vector that contains their ids.  
     * The vector containing forbidding policies' ids is stored in the KAoSSecurityException.
     *
     * @param  aid            	ActionInstanceDescription to be checked by the Guard in order to allow/disallow
     *                         	an action. 
     *
     * @param pObserver        	PolicyDecisionObserver, an entity interested in receiving updates
	 *  						whenever policy decisions change.
	 *                               
     * @throws                	NullPointerException if the aid argument is null.
     *
     * @throws                	ServiceFailure if the policy service is not available, or fails.
     * 
     * @throws					KAoSSecurityException if the action is not permitted.
     * 				
     */
	public void checkDeepPermission (ActionInstanceDescription aid,
									 PolicyDecisionObserver pObserver)
	throws NullPointerException, ServiceFailure, KAoSSecurityException
	{
		if ( aid == null ) throw new NullPointerException("The ActionInstanceDescription is null.");                                                           
      
        String actorID = (String)aid.getActor();
        Guard guard = GuardRetriever.getGuard();
        if ( guard != null ) {
        	guard.checkPermission(new ActionPermission(actorID, aid), null);
        	return;
        }
        
        if ( _dirProxy != null ) {
        	_dirProxy.checkPermission(new ActionPermission(actorID, aid), null);
        	return;
        }
        throw new ServiceFailure("Policy service unavailable! Unable to check permission for: " + aid);
	    
	}
	
	/**
     * The method checks if the given action is permitted according to the current set of policies,
     * or if the given action is a trigger for some obligation policy.
     * 
     * If the action is not permitted, the method throws an exception, but if
     * the action is a trigger, the method returns a Vector containing all
     * those obligation policies for which the action is a trigger.
     *
     * @param  aid              ActionInstanceDescription to be checked by the Guard in order to allow/disallow
     *                          the action; or the trigger to be checked for obligation policies.
     *
     * @param pObserver        	PolicyDecisionObserver, an entity interested in receiving updates
	 *  						whenever policy decisions change.
	 *                                  
     * @throws                	ObligationViolationException some aspect of the returned obligations
     * 							have been violated
     *
     * @throws                	ServiceFailure if the policy service is not available, or fails.
     * 
     * @throws					KAoSSecurityException if the action is not permitted.
     * 				
     */
	public void findPolicyDecision (ActionInstanceDescription aid,
									PolicyDecisionObserver pObserver)
	throws ObligationViolationException, KAoSSecurityException, ServiceFailure
	{
		String actorID = (String)aid.getActor();
	
		Guard guard = GuardRetriever.getGuard();
	    if ( guard != null ) {
	    	guard.findPolicyDecision(new ActionPermission(actorID, aid), null);
	    	return;
	     }
	    if ( _dirProxy != null ) {
	    	_dirProxy.findPolicyDecision(new ActionPermission(actorID, aid), null);
	    	return;
	    }
	    throw new ServiceFailure("Policy service unavailable! Unable to find policy decision for: " + aid);
	}
	
	/**
     * Get policies for the specified action type.
     *
     * @param actionType           	String specifying the action type for the requested policies.  
     *
     * @param pObserver        		PolicyDecisionObserver, an entity interested in receiving updates
	 *  							whenever policy decisions change.
	 *  
     * @return List                 Contains policies defined for the specified action.
     *                        
     */
    public List getPoliciesForActionType (String actionType,
    									  PolicyDecisionObserver pObserver) 
        throws ServiceFailure
    {
        Guard guard = GuardRetriever.getGuard();
        if ( guard == null ) throw new ServiceFailure("Unable to obtain Guard!");
        return guard.getPoliciesForActionType(actionType);  
    }
	
    /**
     * Return policies for the entity, whose id is specified.
     * 
     * @param id      String containing the id of the entity, whose policies are being requested.
     * 
     * @return Vector of PolicyMsg
     */
    public Vector getPoliciesForEntity (String id) throws ServiceFailure
    {
        if ( _dirProxy != null ) {
           return  _dirProxy.getPoliciesForEntity(id);
        }
        
       throw new ServiceFailure("Policy Service unavailable! Unable to obtain policies for: " + id);
    }
    
    
	/**
     * This function is used when the agent/enforcer has only  partial
     * information about an action and would like to determine what range
     * of properties can be allowed by the policy set.
     *
     * The agent/enforcer partially fills an ActionInstanceDescription object
	 * and sends it to the method, which finds those policies that 
	 * are applicable to this action and contain the given property.
	 * The method will then select only those values for the given properties 
	 * that will not conflict with higher priority policies containing
	 * the given properties.
     *      
     * @param  propertyNames    Vector containing the properties for which
	 *						    values are to be found.
	 *  
	 * @param  actionDesc       ActionInstanceDescription object which will
	 *							be used to find applicable policies.
	 *
	 * @param  tight            boolean, if set to 'true', will result in returning
	 *                          only these values for the missing specified property
	 *                          which would satisfy some policy if used alone.
     *
     *  @param pObserver        PolicyDecisionObserver, an entity interested in receiving updates
	 *  						whenever policy decisions change.
	 *  
	 *	@return	Vector			The Vector containing multiple ActionInstanceDescription
	 *							objects which contain the allowed values
	 *							for those properties.
     *
     */
	public Vector getAllowableValuesForActionProperties (Vector propertyNames,
			                                             ActionInstanceDescription actionDesc,
			                                             boolean tight,
			                                             PolicyDecisionObserver pObserver) 
	    throws ServiceFailure
	    {
	        Guard guard = GuardRetriever.getGuard();
            if ( guard == null ) throw new ServiceFailure("Unable to obtain Guard!");
            return guard.getAllowableValuesForActionProperties(propertyNames,actionDesc,tight); 
            
	    }
	
	/** 
	 *  In this case, there are NO properties specified. 
	 *  
	 *	The function uses the given, partially filled, ActionInstanceDescription object to 
	 *  find those policies that are applicable to this action, and then it selects all the 
	 *  properties from those policies which are not originally contained in the provided
	 * 	ActionInstanceDescription object. 
	 *  It then chooses only those values for the given properties which do not conflict 
	 *  with their higher priority selves, and returns all those properties in the 
	 *  ActionInstanceDescription object.
	 * 
	 *  @param  actionDesc		ActionInstanceDescription object which will
	 *							be used to find applicable policies.
	 *
	 *  @param pObserver        PolicyDecisionObserver, an entity interested in receiving updates
	 *  						whenever policy decisions change.
	 * 
	 *	@return	Vector			The Vector containing multiple ActionInstanceDescription
	 *							objects which contain the allowed values
	 *							for those properties.
	 */
	public Vector getAllowableValuesForActionProperties (ActionInstanceDescription actionDesc,
			    										 PolicyDecisionObserver pObserver) 
	 throws ServiceFailure
	 {
	    Guard guard = GuardRetriever.getGuard();
        if ( guard == null ) throw new ServiceFailure("Unable to obtain Guard!");
        return guard.getAllowableValuesForActionProperties(actionDesc); 
	 }

    
    /**
     * This function is used when the enforcer has only  partial
     * information about an action and needs to determine what range
     * of a property can be allowed by the policy.
     *
     * The enforcer partially fills an ActionInstanceDescription object
	 * and sends it to the PolicyDisclosure object to find those policies that 
	 * are applicable to this action and contain the given property.
	 * PolicyDisclosure will then select only those values for the given property 
	 * that will not conflict with higher priority policies containing
	 * the given property.
     *
     * This is a complex function and needs to be understood in the
     * context of how it is used - this is getting written elsewhere
     * and will be included here at a later time.
     *      
     * @param  origPropertyName String specifying the property for which
	 *						    values are to be found.
	 *  
	 * @param  origActionDesc   ActionInstanceDescription object which will
	 *							be used to find applicable policies.
	 
	 * @param  allTargetValues  Set of possible property values - now it is a fake argument,
	 *                          which should really be calculated by the directory service
	 *                          and passed to the entity disclosing the policy (PolicyDisclosure).
	 *
	 * @param  tight            boolean, if set to 'true', will result in returning
	 *                          only these values for the missing specified property
	 *                          which would satisfy some policy if used alone.
     *
     * @param pObserver         PolicyDecisionObserver, an entity interested in receiving updates
	 *  						whenever policy decisions change.
	 *  
	 * @return	Set		        The Set containing allowed values for the given property,
	 *							or null, if a thread waiting to complete execution
	 *                          of this method has been interrupted.  
	 *                          The interrupted status will be propagated.
     * @throws ServiceFailure                          
     */
    public Set getAllowableValuesForActionProperty(String origPropertyName,
                                                   ActionInstanceDescription origActionDesc,
                                                   Set allTargetValues,
                                                   boolean tight,
                                                   PolicyDecisionObserver pObserver)
      throws ServiceFailure
    {
        Guard guard = GuardRetriever.getGuard();
        if ( guard == null ) throw new ServiceFailure("Unable to obtain Guard!");
        return guard.getAllowableValuesForActionProperty(origPropertyName,
                                                         origActionDesc,
                                                         allTargetValues,
                                                         tight); 
    }
    
   /** 
    * Based on the specified trigger ActionInstanceDescription, select all matching control BasicActionDescriptions, which
    * define the obligations for the trigger.  For each control BasicActionDescription create an ActionInstanceDescription.
    * Sort the ActionInstanceDescriptions in descending order by the priority of the obligation policy, in which 
    * the BasicActionDescription was defined.
    *
    * @param triggerAID     ActionInstanceDescription describes an instance of the trigger condition sent
    *                       by the trigger condition monitor.
    *
    * @param pObserver		PolicyDecisionObserver, an entity interested in receiving updates
	*  						whenever policy decisions change.
	*  
    * @return               A Vector of control ActionInstanceDescriptions, whose BasicActionDescriptions are contained 
    *                       in policies matching the parameter triggerAID. The returned control ActionInstanceDescriptions 
    *                       are sorted in descending sequence by their policy priority.
	*   
    * @throws               ObligationViolationException when specified obligation constraints in a policy applicable 
    *                       to the given trigger are not satisfied.
    *
    * @throws               ServiceFailure if the policy service is not available.
    */
   public Vector getObligationsForTriggerCondition(ActionInstanceDescription triggerAID,
		      									   PolicyDecisionObserver pObserver) 
    throws ObligationViolationException, ServiceFailure
    {
        Guard guard = GuardRetriever.getGuard();
        if ( guard != null ) {
            return guard.getObligationsForTriggerCondition(triggerAID);
        }
       throw new ServiceFailure("Policy service unavailable! Unable to retrieve obligations for: " + triggerAID);
    }
    
    /** At the moment this method would work only for A- policies for 'moveTo'
	 *  action, and for 'movingTo' property. For other policies, the behaviour
	 *  of this method is undefined. This method would be made more generic
	 *  in future, for handling nested property values.
	 * 
	 *  @param		origPropertyName		the property for which to find those values.
	 * 
	 *  @param		origActionDesc			the AID for which to find the values.
	 *
     *  @param 		pObserver        		PolicyDecisionObserver, an entity interested in receiving updates
	 *  								    whenever policy decisions change.
	 *  
	 *	@return		Vector					a Vector of OntInstanceDescriptionImpl objects,
	 *										which are the forbidden values.	
	 *  
	 */
	
	public Vector getForbiddenValuesForActionProperty(String origPropertyName,
													  ActionInstanceDescription origActionDesc,
													  PolicyDecisionObserver pObserver) throws ServiceFailure
	{
	    Guard guard = GuardRetriever.getGuard();
        if ( guard != null ) {
            
            return guard.getForbiddenValuesForActionProperty(origPropertyName, origActionDesc);
        }
        throw new ServiceFailure("Policy service unavailable! Unable to obtain forbidden values for property: " 
                                 + origPropertyName);
	}
	
    //NOTE: this method added temporarily here to support a demo!  It will be added 
    // to the implementation of PolicyManagement - where it belongs!!
    /**
    */
    public  void removePolicy (PolicyMsg policyMsg) throws ServiceFailure
    {
        if ( policyMsg != null ) {
            if ( _dirProxy != null ) {
                ArrayList policyList = new ArrayList();
                policyList.add (0, policyMsg);
                //  _kDirService.removePolicies(policyList );
                try {
                    _dirProxy.updatePolicies (new ArrayList(),
                                              new ArrayList(),
                                              policyList) ;
                }
                catch(Exception e){
                    e.printStackTrace();
                    throw new ServiceFailure();
                }
            }
            else {
                throw new ServiceFailure("Policy Service unavailable! Unable to remove policy: " + policyMsg);
            }
        }

    }
    
   
													  
    /*****************************************************************************************
       kaos.core.csi.policy.PolicyChecking interface implementation - end
    *****************************************************************************************/
    /////////////////////////////   PRIVATE METHODS BEGIN
    /*
     **/
    private void initialize(String transportName)
    throws Exception
    {
        try {
            _dirProxy = KAoSBootstrapper.getKAoSDirectoryProxyForTransport(transportName);
        }
        catch(Exception e) {
            e.printStackTrace();
            throw new Exception("    \n\n    *** RegistrationImpl: unable to obtain a reference the DirectoryService for transport: "
                                + transportName + "!");
        }

    }
    
    private ActionInstanceDescription getActionInstanceDescription(String actorID,
                                                                   String actionName,
                                                                   HashMap properties)
    {
        try {
            ActionInstanceDescription returnedAID = new ActionInstanceDescriptionImpl(null, 
            		                                                                  actionName,
            		                                                                  actorID,
            		                                                                  properties);
                    
            return returnedAID;
        } catch (URISyntaxException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        
        return null;
    }
    
    // Utility methods ///////////////////////////////////////////
      
    protected void logMessage(String msg, int logLevel)
    {
   		if (_logger.isEnabledFor(logLevel))
   			_logger.logMessage(msg, logLevel);
    }
	
    protected void logMessage(String msg, Exception e, int logLevel)
    {
   		if (_logger.isEnabledFor(logLevel))
   			_logger.logMessage(msg, e, logLevel);
    }
}
