/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2005/11/28 21:59:11 $
 */
 
package kaos.core.csi.query;

import kaos.core.service.boot.KAoSBootstrapper;
import kaos.core.service.directory.DomainDescription;
import kaos.core.service.directory.DefaultKAoSAgentDescription;
import kaos.core.service.directory.KAoSAgentDescription;
import kaos.core.service.directory.KAoSEntityDescription;
import kaos.core.service.directory.KAoSAgentDirectoryServiceProxy;
import kaos.ontology.util.QueryFailure;
import kaos.core.util.Logger;
import kaos.core.util.NamespaceValidator;
import kaos.ontology.vocabulary.ActorConcepts;

import javax.agent.service.directory.AgentDescription;
import javax.agent.service.directory.DirectoryFailure;
import javax.agent.service.directory.SearchException;

import java.util.List;
import java.util.Map;
import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;

/**
 * Defines the operations to query the different aspects of domain and actor registrations
 * 
 * @author KAoS Team
 * $Revision: 1.11 $
 */
public class QueryRegistrationImpl
    implements QueryRegistration
{
    private static final QueryRegistrationImpl __INSTANCE = new QueryRegistrationImpl() ;
    private static boolean __INITIALIZED = false;
        
    private KAoSAgentDirectoryServiceProxy      _dirProxy;
	private Logger                              _logger = Logger.getLogger(this); 
	
	/**
     * Retrieve an instance of Registration.
     *
     * @param transportName String specifying the name of the transport to use for communication with the KAoS Directory Service.
     * 
     * @return a singleton, Registration.
     */
    public static QueryRegistration getInstance(String transportName)
    throws Exception
    {
        if ( __INITIALIZED ) return __INSTANCE;
        
            __INSTANCE.initialize(transportName);
            __INITIALIZED = true;
            return __INSTANCE;
    }
    
    private QueryRegistrationImpl()
    {
    }

    /*****************************************************************************************
       kaos.core.csi.query.QueryRegistration interface implementation - begin
    *****************************************************************************************/
    /**
     * Returns description of the domain identified by the given name
     *
     * @param name             String specifying the name of the domain to lookup
     *
     * @return                 DomainDescription identified by the given name
     *
     * @throws                 DirectoryFailure if the connection to the DS cannot be established.
     *
     * @throws                 DirectoryFailure if the domain with the given name has not been registered in the DS
     */
    public DomainDescription getDomain(String name) 
        throws DirectoryFailure
    {
        name = NamespaceValidator.validateURI(name);
        DomainDescription desc = _dirProxy.getDomainDescription(name);
        if ( desc == null ) throw new DirectoryFailure();
        return desc;
    }
    
    /**
     * Returns description of the actor identified by the given name
     *
     * @param actorGUID        String specifying the GUID of the actor to lookup
     *
     * @return                 KAoSAgentDescription identified by the given name
     *
     * @throws                 DirectoryFailure if the connection to the DS cannot be established.
     *
     * @throws                 SearchException if the actor with the given name has not been registered in the DS
     */
    public KAoSAgentDescription  getActor(String actorGUID) 
        throws DirectoryFailure, SearchException
    {
        
        actorGUID = NamespaceValidator.validateURI(actorGUID);
        KAoSAgentDescription desc = null;
        desc =  _dirProxy.getAgentDescriptionForAgentID(actorGUID);
        return desc;
    }

    /**
     * Return all agent ids stored in the KAoS directory.
     *
     * @return List         of all agent unique IDs in the agent directory, or null, if there aren't any.
     *
     * @throws              SearchException if an exception is thrown during the search.
     *
     * @throws              DirectoryFailure if the directory cannot be found/is not available.
     */    
    public List getAllAgents()
        throws SearchException, DirectoryFailure
    {
        List agentDescList = _dirProxy.getAllAgentsInDirectory();
        if ( agentDescList == null ) return null;
        Vector agentIDs = new Vector(agentDescList.size());
        for (Iterator agentDescListIt = agentDescList.iterator(); agentDescListIt.hasNext(); ) {
            KAoSAgentDescription agentDesc = (KAoSAgentDescription)agentDescListIt.next(); 
           
            /*
            Locator[] receiverLocs = agentDesc.getLocators();
            if ( receiverLocs.length == 0 ) continue;
            */
            Map ontTypes  = agentDesc.getEntityOntologicalTypes();
            if ( ontTypes.containsKey(ActorConcepts.DomainManager()) ) continue;
            
            String agentID = agentDesc.getEntityNameAsString();
            if ( agentID != null ) {
                agentIDs.add(agentID);
             }
        }
        if (_logger.isEnabledFor(Logger.INFO))
            logMessage(".getAllAgents: returning \n" + agentIDs, Logger.INFO); 
        return agentIDs;
    }
    
      
    /**
     * Returns description of the actors fullfiling constraints expressed as the <a href="http://www-ksl.stanford.edu/knowledge-sharing/kif/">KIF</a> statemant
     *
     * @param actorConstraints String specifying the GUID of the actor to lookup
     * An example of a Constraints statement: 
     *                         Give me all agents that have as capability 'take picture' with a cost of 
     *                         less than $1,- with a reliability of 75% or better and where the average number 
     *                         of processed messages in the past hour is less than 20.
     *
     * @return                 KAoSAgentDescription identified by the given name
     *
     * @throws                 DirectoryFailure if the connection to the DS cannot be established.
     *
     * @throws                 NoSuchActorException  if no actors matching the query properties could be found
     * 
     * @throws                 QueryFailure if query failed
     * 
     */
    public List  getActorsFulfilling(String actorConstraints) 
        throws DirectoryFailure, NoSuchActorException, QueryFailure
    {
        return new ArrayList();
    }

    /**
     * Returns descriptions of the actors whose KAoSAgentDescription contains all specified attributes 
     * and their values.
     *
     * @param attributes        Hashtable containing attribute names and their values to match the attributes in the DS.
     *
     * @return                  List containing  KAoSAgentDescriptions matching the attributes.
     *
     * @throws                  DirectoryFailure if the connection to the DS cannot be established.
     *
     * @throws                 QueryFailure if query failed
     * 
     */
    public List  getActorsWithAttributes(Hashtable attributes) 
        throws DirectoryFailure, QueryFailure
    {
        
        List descList = new ArrayList();
        DefaultKAoSAgentDescription searchDesc =  new DefaultKAoSAgentDescription();
        searchDesc.set(KAoSEntityDescription.SEARCH_PARAMS,attributes);
        
        AgentDescription[] results = null;
        
        try {
            results = _dirProxy.search(searchDesc);
        }
        catch(DirectoryFailure df){
            throw df;
        }
        catch(SearchException se) {
            throw new QueryFailure(se.getMessage());
        }
        
        if ( ( results == null ) || results.length == 0 ) return descList;
        
        for (int i = 0; i < results.length; i++){
            if ( results[i] instanceof KAoSAgentDescription) {
                descList.add((KAoSAgentDescription)results[i]);
            }
        }
        if (_logger.isEnabledFor(Logger.INFO))
            logMessage("getActorsWithAttributes: returning List : " + descList, Logger.INFO);
            
        return descList;
    }

    /*****************************************************************************************
       kaos.core.csi.query.QueryRegistration interface implementation - end
    *****************************************************************************************/
    /////////////////////////////   PRIVATE METHODS BEGIN
    /*
     **/
    private void initialize(String transportName)
    throws Exception
    {
        try {
            _dirProxy = KAoSBootstrapper.getKAoSDirectoryProxyForTransport(transportName);
        }
        catch(Exception e) {
            //e.printStackTrace();
            throw new Exception("    \n\n    *** QueryStateImpl: unable to obtain a reference the DirectoryService for transport: "
                                + transportName + "!");
        }
    }
    //////////////////////////// PRIVATE METHODS END
    // Utility methods ///////////////////////////////////////////
      
    protected void logMessage(String msg, int logLevel)
    {
   	    _logger.logMessage(msg, logLevel);
    }
	
    protected void logMessage(String msg, Exception e, int logLevel)
    {
   	    _logger.logMessage(msg, e, logLevel);
    }
 }
