/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2005/11/28 21:59:11 $
 */
 
package kaos.core.csi.transport;

import kaos.core.KAoSLocator;
import kaos.core.csi.CSIFactory;
import kaos.core.csi.query.QueryRegistration;
import kaos.core.csi.policy.KAoSSecurityException;
import kaos.core.csi.policy.enforcement.CommunicationActionEnforcerImpl;

import kaos.core.service.KAoSServiceRoot;
import kaos.core.service.directory.KAoSAgentDescription;
import kaos.core.util.MsgUtils;

import javax.agent.JasBean;
import javax.agent.Locator;
import javax.agent.TransportMessage;
import javax.agent.service.ServiceFailure;
import javax.agent.service.transport.NotLocatableException;
import javax.agent.service.transport.TransportFailure;
import javax.agent.service.transport.MessageListener;
import javax.agent.service.transport.MessageReceiver;

import java.io.Serializable;
import java.util.Hashtable;
import java.util.List;

/**
 *
 * This class enforcers policies for CommunicationAction before sending messages.
 *
 * The class wraps the TransportSupportImpl in order to authorize message sends. It delegates the operations defined 
 * in TransportSupport to TransportSupportImpl, except for the "sendMessageTo" methods.  Before delegating sending of messages,
 * it authorizes message send and, either allows the operation, or throws an exception
 * to forbid the send.
 *
 * @author KAoS Team
 * $Revision: 1.11 $
 */
public class PolicyEnforcedTransportSupportImpl
    extends CommunicationActionEnforcerImpl 
    implements TransportSupport
{
    private static final PolicyEnforcedTransportSupportImpl     __INSTANCE = new PolicyEnforcedTransportSupportImpl() ;
    
    private TransportSupport                    _transportSupport;
    private QueryRegistration                   _query;
    
	/**
     * Retrieve an instance of PolicyEnforcedTransportSupportImpl.
     *
     * @return a singleton, PolicyEnforcedTransportSupportImpl.
     */
    public static TransportSupport getInstance()
    {
        return __INSTANCE;
    }
    
    private PolicyEnforcedTransportSupportImpl()
    {
         _transportSupport = TransportSupportImpl.getInstance();
         initializeVars();
    }

    public Locator createLocator(String name, JasBean agentInitInfo, String transportName)
    {        
        return _transportSupport.createLocator(name, agentInitInfo, transportName);
    }
   
    /** 
     * Bind the specified MessageListener to the given transport.  
     *
     * @param listener           MessageListener to receive messages using the given Locator (retrieved from the 
     *                          given KAoSAgentDescription.
     *
     * @param listenerName       String containing the name of the receiver.
     *
     * @param listenerDesc       KAoSAgentDescription to be used to retrieve the Locator of the MessageListener.
     *
     *
     * @throws                   NotLocatableException if locator is not a local locator.
     *
     * @throws                   TransportFailure if the binding process fails
     *
     */
    public void bindMsgListenerToTransport(MessageListener listener,
                                           String listenerName,
                                           KAoSAgentDescription listenerDesc)
    throws NotLocatableException,
          TransportFailure
    {
        _transportSupport.bindMsgListenerToTransport(listener, listenerName, listenerDesc);
    }
    
    /** 
     * Bind the specified MessageListener to the given Locator.  
     *
     * @param listener           MessageListener to receive messages using the given Locator.
     *
     * @param locator            Locator to be used to receive messages.
     *
     * @param transportName      String indicating which transport to use
     *
     * @throws                   NotLocatableException if locator is not a local locator.
     *
     * @throws                   TransportFailure if the binding process fails
     *
     */
    public void bindMsgListenerToTransport(MessageListener listener,
                                          String listenerName,
                                          Locator locator,
                                          String transportName)
    throws NotLocatableException,
          TransportFailure
    {
        _transportSupport.bindMsgListenerToTransport(listener, listenerName, locator, transportName);
      
    }
    
    /** 
    * Bind entity, whose name and Locator are given to the transport, whose name is specified.
    * As a result of binding, the entity will receive the MessageReceiver in order to
    * get messages from it (MessageReceiver.getMessage()), as opposed to listening for messages via
    * the MessageListener interface.
    *
    * @param receiverName       String containing the name of the entity binding to transport.
    *
    * @param transportName      String indicating which transport to use
    *
    * @return a MessageReceiver, from which to retrieve messages.
    *
    * @throws                   NotLocatableException if locator is not a local locator.
    *
    * @throws                   TransportFailure if the binding process fails
    *
    */
   public MessageReceiver bindToTransport(String receiverName,
                                          Locator receiverLocator,
                                          String transportName)
   throws NotLocatableException,
          TransportFailure
    {
        return  _transportSupport.bindToTransport(receiverName, receiverLocator, transportName);
    }
    
    /** 
     * Check if the given message send is authorized.  If yes, send  the given message content to the receiver.
     * If not, throw an exception.
     *
     * Send the given message content to the receiver, from whom a message was received.  This method is defined
     * to be used for entities, who are replying to the received message.
     *
     * @param senderDesc         KAoSAgentDescription of the sender.  Based on that the sender Locator and name will be
     *                           determined.
     *
     * @param receivedMsg        TransportMessage that was received by the sender.  Based on that the receiver Locator
     *                           and name will be determined.
     *
     * @param content            Serializable content of the message to be sent out..
     *
     *
     * @throws                   NotLocatableException if the receiver cannot be located.
     *
     * @throws                   TransportFailure if transport fails during message send.
     *
     */
    public void sendMessageTo(KAoSAgentDescription senderDesc,
                              TransportMessage receivedMsg,
                              Serializable content)
        throws NotLocatableException,
              TransportFailure
    {
        Locator sender = senderDesc.getLocators()[0];
        if ( ! (sender instanceof  KAoSLocator) ) 
            throw new NotLocatableException(" Locator not instance of KAoSLocator!  Unable to send the following message: \n"
                                            + content);
                                            
        KAoSLocator loc = (KAoSLocator) sender;
        String transportName = loc.getTransportName();
        
        try {
            authorizeCommunicationAction(senderDesc.getEntityNameAsString(),
                                         getEntityGUID(MsgUtils.getSender(receivedMsg), transportName), 
                                         null,
                                         null,
                                         transportName);
        }
        catch(KAoSSecurityException kse){ 
            kse.printStackTrace();
            throw new TransportFailure("\n\n  !!!! TRANSPORT FAILURE DUE TO SECURITY VIOLATION!  MESSAGE SEND NOT ALLOWED!\n\n");
        }
        catch(NullPointerException npe){
            npe.printStackTrace();
            throw new TransportFailure("\n\n  !!!! TRANSPORT FAILURE DUE TO INCOMPLETE SENDER/RECEIVER INFO!\n\n");
        }
        catch(ServiceFailure sfe){ 
            sfe.printStackTrace();
            throw new TransportFailure("\n\n  !!!! TRANSPORT FAILURE DUE TO POLICY SERVICE FAILURE! MESSAGE SEND NOT ALLOWED!\n\n");
        }
                 
        _transportSupport.sendMessageTo(senderDesc, 
                                        receivedMsg,
                                        content);
                                            
    }
    
   /** 
     * Check if the given message send is authorized.  If yes, send  the given message content to the receiver.
     * If not, throw an exception.
     *
     * Send the given message content to the receiver, whose KAoSAgentDescription is given.  This method is defined
     * to be used for entities looking-up other entities in the Directory Service in order to send messages.
     *
     * @param senderDesc         KAoSAgentDescription of the sender.  Based on that the sender Locator and name will be
     *                           determined.
     *
     * @param receiverDesc       KAoSAgentDescription of the receiver.  Based on that the receiver Locator and name will be
     *                           determined.
     * 
     * @param content            Serializable content of the message to be sent out.
     *
     *
     * @throws                   NotLocatableException if the receiver cannot be located.
     *
     * @throws                   TransportFailure if transport fails during message send.
     *
     */
    public void sendMessageTo(KAoSAgentDescription senderDesc,
                             KAoSAgentDescription receiverDesc,
                             Serializable content)
        throws NotLocatableException,
            TransportFailure
    {
        Locator sender = senderDesc.getLocators()[0];
        if ( ! (sender instanceof  KAoSLocator) ) 
            throw new NotLocatableException(" Locator not instance of KAoSLocator!  Unable to send the following message: \n"
                                            + content);
        
        KAoSLocator loc = (KAoSLocator) sender;
        String transportName = loc.getTransportName();
        
        try {
            authorizeCommunicationAction(senderDesc.getEntityNameAsString(),
                                         receiverDesc.getEntityNameAsString(),
                                         null,
                                         null,
                                         transportName);
        }
        catch(KAoSSecurityException kse){ 
            kse.printStackTrace();
            throw new TransportFailure("\n\n  !!!! TRANSPORT FAILURE DUE TO SECURITY VIOLATION!  MESSAGE SEND NOT ALLOWED!\n\n");
        }
        catch(NullPointerException npe){
            npe.printStackTrace();
            throw new TransportFailure("\n\n  !!!! TRANSPORT FAILURE DUE TO INCOMPLETE SENDER/RECEIVER INFO!\n\n");
        }
        catch(ServiceFailure sfe){ 
            sfe.printStackTrace();
            throw new TransportFailure("\n\n  !!!! TRANSPORT FAILURE DUE TO POLICY SERVICE FAILURE! MESSAGE SEND NOT ALLOWED!\n\n");
        }
        
        _transportSupport.sendMessageTo(senderDesc, 
                                        receiverDesc,
                                        content);
    }
    
   /** 
    * Check if the given message send is authorized.  If yes, send  the given message content to the receiver, whose Locator is given.
    * If not, throw an exception.
    *
    * @param senderLocator      Locator of the sender.
    *
    * @param receiverLocator    Locator of the receiver.
    *
    * @param senderName         String containing the name of the sender.
    *
    * @param receiverName       String containing the name of the receiver.
    *
    * @param fipaPerformative   String containing the FIPA performative describing the message intent.    
    *
    * @param content            Serializable containing the message content.    
    *
    * @param transportName      String containing the name of the transport to be used.    
    *
    * @throws                   NotLocatableException if the receiver cannot be located.
    *
    * @throws                   TransportFailure if transport fails during message send.
    *
    */
    public void sendMessageTo(Locator senderLocator,
                              Locator receiverLocator,
                              String senderName,
                              String receiverName,
                              String fipaPerformative,
                              Serializable content,
                              String transportName)
    throws NotLocatableException,
          TransportFailure
    {
        if ( transportName == null ) throw new NotLocatableException(" Transport name is null, message: " + content + "\n cannot be sent out!");
        
        try {
            authorizeCommunicationAction(getEntityGUID(senderName,transportName),
                                         getEntityGUID(receiverName,transportName),
                                         null,
                                         null,
                                         transportName);
        }
        catch(KAoSSecurityException kse){ 
            kse.printStackTrace();
            throw new TransportFailure("\n\n  !!!! TRANSPORT FAILURE DUE TO SECURITY VIOLATION!  MESSAGE SEND NOT ALLOWED!\n\n");
        }
        catch(NullPointerException npe){
            npe.printStackTrace();
            throw new TransportFailure("\n\n  !!!! TRANSPORT FAILURE DUE TO INCOMPLETE SENDER/RECEIVER INFO!\n\n");
        }
        catch(ServiceFailure sfe){ 
            sfe.printStackTrace();
            throw new TransportFailure("\n\n  !!!! TRANSPORT FAILURE DUE TO POLICY SERVICE FAILURE! MESSAGE SEND NOT ALLOWED!\n\n");
        }
        
        
       _transportSupport.sendMessageTo(senderLocator, 
                                       receiverLocator,
                                       senderName,
                                       receiverName,
                                       fipaPerformative,
                                       content,
                                       transportName);
    }
    
    /** 
     * Check if the given message send is authorized.  If yes, send  the given message content to the receiver, whose Locator is given.
     * If not, throw an exception.
     *
     * @param senderLocator      Locator of the sender.
     *
     * @param receiverLocator    Locator of the receiver.
     *
     * @param senderName         String containing the name of the sender.
     *
     * @param receiverName       String containing the name of the receiver.
     *
     * @param fipaPerformative   String containing the FIPA performative describing the message intent.    
     *
     * @param content            Serializable containing the message content.    
     *
     * @throws                   NotLocatableException if the receiver cannot be located.
     *
     * @throws                   TransportFailure if transport fails during message send.
     *
     */
    public void sendMessageTo(Locator senderLocator,
                              Locator receiverLocator,
                              String senderName,
                              String receiverName,
                              String fipaPerformative,
                              Serializable content)
    throws NotLocatableException,
          TransportFailure
    {
        if ( ! (receiverLocator instanceof  KAoSLocator) ) 
            throw new NotLocatableException("Receiver Locator not instance of KAoSLocator!  Unable to send a message to: "
                                            + "\n   Receiver name: receiverName" + "\n   Receiver locator: " +  receiverLocator
                                            + "\n   Message to send: " + content);
                                            
        KAoSLocator loc = (KAoSLocator) receiverLocator;
        String transportName = loc.getTransportName();
        
        try {
            authorizeCommunicationAction(getEntityGUID(senderName,transportName),
                                         getEntityGUID(receiverName,transportName),
                                         null,
                                         null,
                                         transportName);
        }
        catch(KAoSSecurityException kse){ 
            kse.printStackTrace();
            throw new TransportFailure("\n\n  !!!! TRANSPORT FAILURE DUE TO SECURITY VIOLATION!  MESSAGE SEND NOT ALLOWED!\n\n");
        }
        catch(NullPointerException npe){
            npe.printStackTrace();
            throw new TransportFailure("\n\n  !!!! TRANSPORT FAILURE DUE TO INCOMPLETE SENDER/RECEIVER INFO!\n\n");
        }
        catch(ServiceFailure sfe){ 
            sfe.printStackTrace();
            throw new TransportFailure("\n\n  !!!! TRANSPORT FAILURE DUE TO POLICY SERVICE FAILURE! MESSAGE SEND NOT ALLOWED!\n\n");
        }
        
        
        _transportSupport.sendMessageTo(senderLocator, 
                                        receiverLocator,
                                        senderName,
                                        receiverName,
                                        fipaPerformative,
                                        content);
    }

    /** 
     * Return the names of the currently supported transports.
     *
     * @return a List containing Strings representing names of supported transports.
     *
     **/
    public List getSupportedTransportsNames()
    {
        return _transportSupport.getSupportedTransportsNames();
    }
   
   /** 
    * Answer true if the given transport is supported, otherwise, answer false.
    *
    * @param name       String containing the name of the transport in question.
    *
    * @return a boolean indicating if the given transport is supported.
    *
    **/
    public boolean isTransportSupported(String name)
    {
        return _transportSupport.isTransportSupported(name);
    }
   
   /** 
    * Return  KAoSServiceRoot for the specified transportName.
    *
    * @param transportName          String containing the name of the transport to create a KAoSServiceRoot for.
    *
    * @return a KAoSServiceRoot     specific for the transportName.
    *
    **/
    public KAoSServiceRoot getServiceRoot(String transportName)
    {
        return _transportSupport.getServiceRoot(transportName);
    }
    
    
    // PRIVATE METHODS
    
    private String getEntityGUID(String nickname, String dsTransportName)
    {
        if ( nickname == null || dsTransportName == null ) return "UNKNOWN";
        
        if ( _query == null ) {
            try {
                CSIFactory.setCurrentTransport(dsTransportName);
                _query = CSIFactory.getQueryRegistration();
            }
            catch(Exception e){
                e.printStackTrace();
                return "UNKNOWN";
            }
        }
        
        Hashtable searchHash = new Hashtable();
        searchHash.put(KAoSAgentDescription.AGENT_NICKNAME,nickname);
        try {
            List descriptions  = _query.getActorsWithAttributes(searchHash);
            if ( descriptions != null &&  ( descriptions.size() > 0 ) ){
                KAoSAgentDescription desc = (KAoSAgentDescription) descriptions.get(0);
                return desc.getEntityNameAsString();
            }
            else {
                return "UNKNOWN";
            }
        }
        catch(Exception e){
            e.printStackTrace();
            return "UNKNOWN";
        }
    }
    
 }
