/**
 * Last Modified by: $Author: auszok $
 * On: $Date: 2005/11/28 15:32:56 $
 */
package kaos.core.csi.usecase;

import kaos.core.csi.request.ActionMethodRequestHandler;
import kaos.core.service.directory.DefaultDomainDescription;
import kaos.core.csi.CSIFactory;
import kaos.core.csi.registration.Registration;
import kaos.core.service.ontology.request.ActionStatusCallback;
import kaos.ontology.vocabulary.ActionConcepts;
import kaos.ontology.vocabulary.ActorConcepts;
import kaos.policy.guard.GuardInstantiationException;

import javax.agent.service.directory.AlreadyRegisteredException;
import javax.agent.service.directory.DirectoryFailure;

import java.util.GregorianCalendar;
import java.util.StringTokenizer;
import java.util.HashMap;
import java.util.Enumeration;
import java.util.Vector;

/**
 * $Revision: 1.4 $
 */
public class RequestReceiverUsecase
    extends ActionMethodRequestHandler
    
 {   
    /**
        This class llustrates an implementation of a CSI RequestReceiver: an entity capable of receiving and processing requests for 
        actions defined in kaos ontologies.
        
        Configuration data includes: nickname, names of kaos domains it will register in, name of a transport 
        to use for communication within KAoS.
        
        Implementation:
	        - inherits from a generic CSI RequestReceiver,  which also binds to transport and policy service
	        - adds specific requests for CommunicationAction()
	            async with and without properties
	            sync with and without properties
	        - each received request is authorized by the CSI Policy
	        
        Bootstrapping steps:
	        - obtains a reference to CSI Registration
	        - binds itself to CSI Transport (implemented in its superclass)
	        - binds itself to  CSI Policy Service (implemented in its superclass)
	        - asks CSI Registration  to create a KAoSAgentDescription for  itself
	        - specifies ontological type of a Robot in the KAoSAgentDescription
	        - asks CSI Registration to register with KAoS
	        
        Running
	        - waits for requests for CommunicationAction and processes those requests when they come in.

    */
    
    public RequestReceiverUsecase( )
    {
    }

    
    //////////////////////////////////////////////////////////////

    public void initialize(String name, Vector domainNames, String transportName)
    throws Exception
    {
        Registration registration = null;
        
        CSIFactory.setCurrentTransport(transportName);
        registration = CSIFactory.getRegistration();
       
        _agentDescription = registration.createKAoSAgentDescription(null,       // guid - will be assigned by KAoS
                                                                    name,       // nickname
                                                                    domainNames,
                                                                    transportName,
                                                                    null);      // no extra transport info necessary
                                                                    
                                                                    
        _agentDescription.setEntityOntologicalType(ActorConcepts.Robot());
        
        bindToTransport(_agentDescription);
        
        bindToPolicyService(transportName);
        
        registration.registerEntity(_agentDescription,
                                    true,       // guarded
                                    false);     // replace registration  
                  
        System.out.println("\n   RequestReceiverUsecase: registered in KAoS using KAoSAgentDescription:\n " + _agentDescription);
       
    }
    
    // async with properties and callback
    public void CommunicationAction(HashMap properties,
                                    ActionStatusCallback statusCallback) 
    throws Exception
    {
        System.out.println("\n   CommunicationAction[HashMap,ActionStatusCallback]: received properties: " + properties
                            + "\n   statusCallback: " + statusCallback);
                            
        String destination = (String)properties.get(ActionConcepts.hasDestination());
        if ( destination == null ) destination = "UNKNOWN";
        
        System.out.println("\n   CommunicationAction[HashMap,ActionStatusCallback]: will reply via statusCallback.onAccepted:  Accepted CommunicationAction to: " 
                            + destination);
        if ( statusCallback != null ) {
            statusCallback.onAccepted(new GregorianCalendar(), "Accepted CommunicationAction to: " + destination);
            statusCallback.onCompleted(new GregorianCalendar(), "Completed CommunicationAction to: " + destination);
        }
        
    }
    
    // async without properties
    public void CommunicationAction(ActionStatusCallback statusCallback) 
    throws Exception
    {
        System.out.println("\n   CommunicationAction[ActionStatusCallback]: no properties processed. Will reply via statusCallback.onAccepted:  Accepted CommunicationAction WITH NO PROPERTIES!");
                            
        if ( statusCallback != null ) {
            statusCallback.onAccepted(new GregorianCalendar(), "Accepted CommunicationAction WITH NO PROPERTIES!");
            statusCallback.onCompleted(new GregorianCalendar(), "Completed CommunicationAction WITH NO PROPERTIES!");
        }
        
    }
    
    // sync with properties
    public String CommunicationAction(HashMap properties) 
    throws Exception
    {
        System.out.println("\n   CommunicationAction[HashMap]: Received properties:  " + properties);
                            
        String destination = (String)properties.get(ActionConcepts.hasDestination());    
        if ( destination == null ) destination = "UNKNOWN";
        System.out.println("\n   CommunicationAction[HashMap]: Returning: Will communicate to: " + destination);
        return "Will communicate to: " + destination;
    }
    
    // sync without properties
    public String CommunicationAction() 
    throws Exception
    {
        System.out.println("\n   CommunicationAction[]: no args version! Will return:  Will communicate to everyone!");
                            
        return "Will communicate to everyone!";
    }
    
    // sync without properties, returning null
    public String MonitorAction() 
    throws Exception
    {
        System.out.println("\n   MonitorAction[]: no args version! Will return null!");
                            
        return null;
    }
    
   //////////////////////////////////////////////////////////////
   // Entry point...

   public static void main(String [] args)
      throws Exception      
    {
        String name = null;
        String transport = null;
        Vector domainNames = null;
        if (args.length > 0) {
	        int i = 0;
	        while (i < args.length) {
	            if (args[i].equalsIgnoreCase ("-name")) {
	                if ((i+1) < args.length) {
    	                name = args[i+1];
    	                i++;
    	            }
    	            else {
    	                System.out.println ("Must specify Agent Name with -name");
    	                System.exit(0);
    	            }
	            }
	            else if (args[i].equalsIgnoreCase ("-transport")) {
	                if ((i+1) < args.length) {
    	                transport = args[i+1];
    	                i++;
    	            }
    	            else {
    	                System.out.println ("Must specify  transport with -transport.  Valid transports: corba, grid, tcp.");
    	                System.exit(0);
    	            }
	            }
	            else if (args[i].equalsIgnoreCase ("-domains")) {
                    if ((i+1) < args.length) {
                        String domainsStr = args[i+1];
                        domainNames = new Vector();
                        StringTokenizer tokenizer = new StringTokenizer(domainsStr, ",");
                        while (tokenizer.hasMoreTokens()) {
                            domainNames.add(tokenizer.nextToken());
                        }
                    }
                }// else if
                 i++;
	        }// while
	    }
	    
        if ( name == null) {
            System.out.println ("\n     *** Must specify Agent Name with -name\n");
            return;
        }
        
        if ( transport == null) {
            System.out.println ("\n     *** Must specify  transport with -transport.  Valid transports: corba, grid, tcp.");
            return;
        }
        
        if ( domainNames == null ) {
            domainNames.add("TestDomain");
            System.out.println (" <>    Warning: No Domain Names specified with -domains, using default: " + domainNames);
        }
        
        System.out.println ("\n     <>  Name of agent to register:  " + name);
	    System.out.println ("\n     <>  Requested transport:        " + transport);
	    System.out.println ("\n     <>  Requested domains:          " + domainNames + "\n");
	    	    
        System.out.println ("\n Will request RegistrationImpl  using transport: " + transport);
        
        Registration registration = null;
        try {
            CSIFactory.setCurrentTransport(transport);
            registration = CSIFactory.getRegistration();
        }
        catch(Exception e){
            e.printStackTrace();
            System.out.println ("  **** Unable to obtain CSI services: Registration. Exiting...");
            return;
        }
        
        System.out.println ("\n Obtained registration " + registration + "\n\n");
        Enumeration domainsEnum = domainNames.elements();
        while ( domainsEnum.hasMoreElements() ) {
            String domainName  = (String) domainsEnum.nextElement();
            DefaultDomainDescription desc = new DefaultDomainDescription();
            desc.setDomainName(domainName);
            System.out.println("\n  Will register domain: " + desc + "\n");
            
            try {
                registration.registerEntity(desc,
                                            false,      // guarded?
                                            false);     // replace existing registration?
            }
            catch(DirectoryFailure de) {
                System.out.println("\n   Due to Directory Failure unable to register domain: " + domainName);
                de.printStackTrace();
                return;
            }
            catch(AlreadyRegisteredException ae) {
                System.out.println("    The following domain has already been registered: " + domainName);
            }
            catch(GuardInstantiationException gie) {
                // this should not happen...
            }
            
        }
        
        RequestReceiverUsecase usecase = new RequestReceiverUsecase();
        
        try {
            usecase.initialize(name, domainNames, transport);
        }
        catch(Exception e) {
            e.printStackTrace();
            return;
        }
       
        synchronized ( Thread.currentThread() ) {
         Thread.currentThread().wait();

      }

   }
}

